package com.chelvc.framework.common.model;

import java.io.Serializable;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.chelvc.framework.common.util.StringUtils;
import com.fasterxml.jackson.annotation.JsonIgnore;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.NonNull;

/**
 * 地区编码
 *
 * @author Woody
 * @date 2024/1/30
 */
@Getter
@Builder
@EqualsAndHashCode
@NoArgsConstructor
@AllArgsConstructor
public final class Region implements Serializable {
    /**
     * 地区编码最小值
     */
    private static final int MIN_VALUE = 100000;

    /**
     * 地区编码最大值
     */
    private static final int MAX_VALUE = 999999;

    /**
     * 地区编码
     */
    private int code;

    /**
     * 地区边界编码
     */
    private int boundary;

    /**
     * 判断是否是地区编码
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isRegion(int code) {
        return code >= MIN_VALUE && code <= MAX_VALUE;
    }

    /**
     * 判断是否是市级编码
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isCity(int code) {
        return isRegion(code) && (code % 10000 > 0) && (code % 100 == 0);
    }

    /**
     * 判断是否是省级编码
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isProvince(int code) {
        return isRegion(code) && (code % 10000 == 0);
    }

    /**
     * 判断是否是区级编码
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isDistrict(int code) {
        return isRegion(code) && (code % 100 > 0);
    }

    /**
     * 判断地区是否是直辖市
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isDirectly(int code) {
        code = code / 10000;
        return code == 11 || code == 12 || code == 31 || code == 50;
    }

    /**
     * 判断地区是否属于敏感地区
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isSensitive(int code) {
        code = code / 10000;
        return code == 11 || code == 65 || code == 54;
    }

    /**
     * 判断是否是同一地区
     *
     * @param code1 地区编码
     * @param code2 地区编码
     * @return true/false
     */
    public static boolean isSame(int code1, int code2) {
        return isRegion(code1) && isRegion(code2) && (Objects.equals(code1, code2)
                || Objects.equals(code1, code2 / 100 * 100) || Objects.equals(code1, code2 / 10000 * 10000)
                || Objects.equals(code1 / 100 * 100, code2) || Objects.equals(code1 / 10000 * 10000, code2));
    }

    /**
     * 判断是否是同一地区
     *
     * @param region1 地区对象
     * @param region2 地区对象
     * @return true/false
     */
    public static boolean isSame(Region region1, Region region2) {
        return region1 != null && region2 != null && isSame(region1.getCode(), region2.getCode());
    }

    /**
     * 将地区编码转换成市级编码
     *
     * @param code 地区编码
     * @return 地区编码
     */
    public static int code2city(int code) {
        return code / 100 * 100;
    }

    /**
     * 将地区编码转换成省级编码
     *
     * @param code 地区编码
     * @return 地区编码
     */
    public static int code2province(int code) {
        return code / 10000 * 10000;
    }

    /**
     * 构建地区编码
     *
     * @param code 地区编码
     * @return 地区实例
     */
    public static Region of(int code) {
        if (isProvince(code)) {
            return new Region(code, code + 9999);
        } else if (isCity(code)) {
            return new Region(code, code + 99);
        } else if (isDistrict(code)) {
            return new Region(code, 0);
        }
        throw new IllegalArgumentException("Invalid region: " + code);
    }

    /**
     * 获取地区简称
     *
     * @param name 地区名称
     * @return 地区简称
     */
    public static String shortname(String name) {
        if (StringUtils.isEmpty(name)) {
            return name;
        }
        if (name.endsWith("特别行政区")) {
            name = name.substring(0, name.length() - 5);
        }
        if (name.endsWith("地区")) {
            name = name.substring(0, name.length() - 2);
        }
        if (name.endsWith("自治县") || name.endsWith("自治区") || name.endsWith("自治州")
                || name.endsWith("直辖市") || name.endsWith("市辖区")) {
            name = name.substring(0, name.length() - 3);
        }
        if (name.endsWith("省") || name.endsWith("市") || name.endsWith("县")) {
            name = name.substring(0, name.length() - 1);
        }
        return name;
    }

    /**
     * 获取地区简称
     *
     * @param name      地区名称
     * @param delimiter 地区分割符
     * @return 地区简称
     */
    public static String shortname(String name, @NonNull String delimiter) {
        if (StringUtils.isEmpty(name)) {
            return name;
        }
        return Stream.of(name.split(delimiter)).map(Region::shortname).collect(Collectors.joining(StringUtils.EMPTY));
    }

    /**
     * 获取地区上级地区
     *
     * @return 地区信息
     */
    @JsonIgnore
    public Region getParent() {
        if (this.isCity()) {
            int province = code2province(this.code);
            return new Region(province, province + 9999);
        } else if (this.isDistrict()) {
            int city = code2city(this.code);
            return new Region(city, city + 99);
        }
        return null;
    }

    /**
     * 判断是否是市级地区
     *
     * @return true/false
     */
    @JsonIgnore
    public boolean isCity() {
        return isCity(this.code);
    }

    /**
     * 判断是否是省级地区
     *
     * @return true/false
     */
    @JsonIgnore
    public boolean isProvince() {
        return isProvince(this.code);
    }

    /**
     * 判断是否是区级地区
     *
     * @return true/false
     */
    @JsonIgnore
    public boolean isDistrict() {
        return isDistrict(this.code);
    }

    /**
     * 判断是否是直辖市
     *
     * @return true/false
     */
    @JsonIgnore
    public boolean isDirectly() {
        return isDirectly(this.code);
    }

    /**
     * 判断地区是否属于敏感地区
     *
     * @return true/false
     */
    @JsonIgnore
    public boolean isSensitive() {
        return isSensitive(this.code);
    }

    /**
     * 判断是否是同一地区
     *
     * @param code 地区编码
     * @return true/false
     */
    @JsonIgnore
    public boolean isSame(int code) {
        return isSame(this.code, code);
    }

    /**
     * 判断是否是同一地区
     *
     * @param region 地区实例
     * @return true/false
     */
    @JsonIgnore
    public boolean isSame(Region region) {
        return isSame(this, region);
    }

    @Override
    public String toString() {
        return String.valueOf(this.code);
    }
}
